import numpy as np
import pandas as pd
import os
import json
import matplotlib.pyplot as plt
import pickle


class Data(object):
    """Base class for data set"""

    def __init__(self, name, root, config_file, output_dir=None):
        self.name = name
        self.root = root
        self.all_data = None
        self.output_dir = output_dir
        if output_dir is not None and not os.path.exists(output_dir):
            os.makedirs(output_dir)
        config_file = os.path.join(root, config_file)
        with open(config_file) as fh:
            self.config = json.load(fh)
        self.has_test = self.has_test_set()

    def has_test_set(self):
        """ does the dataset contain independent test set """
        if 'test_dataset' in self.config:
            return True
        else:
            return False

    def load_data(self, data_file, samples, vis=False):
        # load feature data
        X = pd.read_csv(data_file, sep='\t', index_col=0)
        # if there are duplicated rows (genes), take the row average
        X = X.groupby(X.index).mean()
        # sample x features
        X = X.T
        if samples is not None:
            X = X.loc[samples, :]
        if vis and self.output_dir is not None:
            myfig = plt.figure()
            X_temp = X.T
            flierprops = dict(marker='o', markersize=1)
            X_temp.boxplot(grid=False, rot=90, fontsize=2,
                flierprops=flierprops)
            upper_dir = os.path.basename(os.path.dirname(data_file))
            file_name = (self.name + '_' + upper_dir + '_' +
                         os.path.basename(data_file) + '.pdf')
            myfig.savefig(os.path.join(self.output_dir, file_name),
                          format='pdf')

        X.sort_index(inplace=True)
        # remove features with missing values
        X.dropna(axis=1, inplace=True)
        return X

    def save_data(self):
        if self.output_dir is not None:
            fname = self.name + '_all_data.pkl'
            with open(os.path.join(self.output_dir, fname), 'wb') as fh:
                pickle.dump(self.all_data, fh, protocol=pickle.HIGHEST_PROTOCOL)


class Dataset(Data):
    """ data set"""

    def __call__(self):
        print('processing data ...')
        train_dataset = self.config['train_dataset']
        target_view = self.config['target_view']
        target_label = self.config['target_label']
        target_view = self.config['target_view']
        clin_file = self.config['data'][train_dataset]['label']['file']
        clin_file = os.path.join(self.root, train_dataset, clin_file)
        # the first col will be the index
        clin_data = pd.read_csv(clin_file, sep='\t', index_col=0)
        # samples X phenotype
        clin_data = clin_data.loc[:, [target_label]]
        train_sample = clin_data.index
        # must be binary value (0 or 1)
        label_val = sorted(np.unique(clin_data.iloc[:, 0].values))
        assert label_val == [0, 1], 'label can only be 0 or 1'
        y = clin_data

        # get the name of samples that have all the omics data available
        all_views = self.config['data'][train_dataset]['view']
        n_views = len(all_views)
        all_view_names = [item['type'] for item in all_views]
        all_view_names.remove(target_view)
        all_view_ordered = [target_view] + sorted(all_view_names)

        if self.has_test:
            test_dataset = self.config['test_dataset']
            all_test_views = self.config['data'][test_dataset]['view']
            if 'label' in self.config['data'][test_dataset]:
                test_clin_file = self.config['data'][test_dataset]['label']['file']
                test_clin_file = os.path.join(self.root, test_dataset, test_clin_file)
                test_clin_data = pd.read_csv(test_clin_file, sep='\t', index_col=0)
                test_samples = test_clin_data.index
                # samples X phenotype
                test_clin_data = test_clin_data.loc[:, [target_label]]
                label_val = sorted(np.unique(test_clin_data.iloc[:, 0].values))
                assert label_val == [0, 1], 'label can only be 0 or 1'
                y_final_test_2 = test_clin_data
            else:
                test_samples = None
                y_final_test_2 = None
            selected_view = filter(lambda view: view['type'] == target_view,
                                   all_test_views)
            test_view_file = list(selected_view)[0]['file']
            test_view_file = os.path.join(self.root, test_dataset, test_view_file)
            X_final_test_2 = self.load_data(test_view_file, test_samples)
        else:
            X_final_test_2 = None
            y_final_test_2 = None

        all_data_ = {'desc': {
                        'name': self.name,
                        'has_test': self.has_test,
                        'target_label': target_label,
                        'target_view': target_view,
                        'n_view': n_views,
                        'view_names': all_view_ordered
                        },
                     'X': {},
                     'y': y,
                     'X_test': None,
                     'y_test': y_final_test_2
                    }

        for i in range(n_views):
            view_name = all_views[i]['type']
            print(f'view: {view_name}')
            view_file = os.path.join(self.root, train_dataset,
                                     all_views[i]['file'])
            X = self.load_data(view_file, train_sample)
            if view_name == target_view:
                if self.has_test:
                    # find out common features in the two data
                    # sets for the target view
                    common_features = sorted(list(set(X.columns.values) & set(
                        X_final_test_2.columns.values)))
                    print('train data target view features length {}'.format(
                        len(X.columns.values)))
                    print('test data target view features length {}'.format(
                        len(X_final_test_2.columns.values)))
                    print('common target view features length {}'.format(
                        len(common_features)))
                    X = X[common_features]
                    X_final_test_2 = X_final_test_2[common_features]
                    X_final_test_2 = X_final_test_2.add_prefix(view_name + '_')

            X = X.add_prefix(view_name + '_')
            all_data_['X'][view_name] = X

        if self.has_test:
            all_data_['X_test'] = X_final_test_2

        print('number of samples: {}'.format(all_data_['y'].shape[0]))

        if self.has_test:
            all_data_['X_test'] = X_final_test_2
            if all_data_['y_test'] is not None:
                print('test sample size: {}'.format(all_data_['y_test'].shape[0]))
            else:
                print('test sample size: {}'.format(all_data_['X_test'].shape[0]))

        self.all_data = all_data_
        self.save_data()
        return all_data_



from sklearn.metrics import roc_auc_score
from sklearn.utils import check_X_y
import numpy as np
import multiprocessing


def _my_auc_score(y_pred, y_true):
    return roc_auc_score(y_true, y_pred)


# https://stackoverflow.com/a/45555516/410069
def unpacking_apply_along_axis(all_args):
    """
    Like numpy.apply_along_axis(), but with arguments in a tuple
    instead.

    This function is useful with multiprocessing.Pool().map(): (1)
    map() only handles functions that take a single argument, and (2)
    this function can generally be imported from a module, as required
    by map().
    """
    (func1d, axis, arr, args, kwargs) = all_args
    return np.apply_along_axis(func1d, axis, arr, *args, **kwargs)


def parallel_apply_along_axis(func1d, axis, arr, *args, **kwargs):
    """
    Like numpy.apply_along_axis(), but takes advantage of multiple
    cores.
    """
    # Effective axis where apply_along_axis() will be applied by each
    # worker (any non-zero axis number would work, so as to allow the use
    # of `np.array_split()`, which is only done on axis 0):
    effective_axis = 1 if axis == 0 else axis
    if effective_axis != axis:
        arr = arr.swapaxes(axis, effective_axis)

    # Chunks for the mapping (only a few chunks):
    chunks = [(func1d, effective_axis, sub_arr, args, kwargs)
              for sub_arr in np.array_split(arr, multiprocessing.cpu_count())]

    pool = multiprocessing.Pool()
    individual_results = pool.map(unpacking_apply_along_axis, chunks)
    # Freeing the workers:
    pool.close()
    pool.join()

    return np.concatenate(individual_results)


def sym_auc_score(X, y):
    """Compute the symmetric auroc score for the provided sample.
    symmetric auroc score is defined as 2*abs(auroc-0.5)

    Parameters
    ----------
    X : {array-like, sparse matrix} shape = [n_samples, n_features]
        The set of regressors that will be tested sequentially.

    y : array of shape(n_samples)
        The data matrix.

    Returns
    -------
    F : array, shape = [n_features,]
        The set of auroc scores.
    """
    X, y = check_X_y(X, y, accept_sparse=['csr', 'csc', 'coo'])
    # scores = parallel_apply_along_axis(_my_auc_score, 0, X, y)
    scores = np.apply_along_axis(_my_auc_score, 0, X, y)
    return np.abs(scores - 0.5) * 2.0



import numpy as np
import pandas as pd
import re
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection import f_classif
from .k_medoids import KMedoids
from .utils import sym_auc_score
from sklearn.feature_selection import f_classif
from sklearn.feature_selection import SelectPercentile
import warnings
from sklearn.feature_selection._univariate_selection import _BaseFilter
from sklearn.feature_selection._univariate_selection import _clean_nans
from sklearn.utils.validation import check_is_fitted


class SelectMinScore(_BaseFilter):
    """
    select features with minimum score threshold
    """

    def __init__(self, score_func=f_classif, min_score=0.5):
        super(SelectMinScore, self).__init__(score_func)
        self.min_score = min_score
        self.score_func = score_func

    def _check_params(self, X, y):
        if not (self.min_score == "all" or 0 <= self.min_score <= 1):
            raise ValueError('min_score should be >=0, <= 1; got {}.'
                             'Use min_score="all" to return all features.'
                             .format(self.min_score))

    def _get_support_mask(self):
        check_is_fitted(self, 'scores_')
        if self.min_score == 'all':
            return np.ones(self.scores_.shape, dtype=bool)

        scores = _clean_nans(self.scores_)
        mask = np.zeros(scores.shape, dtype=bool)
        mask = scores > self.min_score
        # Note: it is possible that mask contains all False.
        return mask


class FeatureSelector(BaseEstimator, TransformerMixin):
    def __init__(self, views, target_view, method, k=3,
                 weighted=False, percentile=1):
        self.views = views
        self.target_view = target_view
        self.method = method
        self.mode = self.get_filter_mode()
        self.k = k
        self.weighted = weighted
        self.percentile = percentile
        self.cluster_membership = None
        self.selected_features = None
        self.data = {}
        self.support = {}
    
    def get_filter_mode(self):
        if self.method.endswith('_mo'):
            return 'mo'
        return 'so'

    def feature_sel(self):
        fs_method = fs_methods[self.method]
        if self.method in ['proms', 'proms_mo']:
            # cluster_membership is a dictionary with selected
            # markers as keys
            ret = fs_method(self.data, self.target_view,
                            self.k, self.weighted)()
            selected_features, cluster_membership = ret
            return (selected_features, cluster_membership)
        else:
            raise ValueError('method {} is not supported'.format(self.method))

    def assemble_data(self, X, y=None):
        ''' X is a combined multi-view data frame
        '''
        for view in self.views:
            # find features in each view
            ptn = re.compile(r'^{}_'.format(view))
            cur_view_features = [i for i in self.all_features
                                 if ptn.match(i)]
            cur_X = X.loc[:, cur_view_features]
            self.data[view] = {}
            self.data[view]['X'] = cur_X
            self.data[view]['y'] = y

    def single_view_prefilter(self, X, y=None):
        view = self.views[0]
        if view != self.target_view:
            raise ValueError('target view name not matched')
        ptn = re.compile(r'^{}_'.format(view))
        cur_view_features = [i for i in self.all_features
                             if ptn.match(i)]
        cur_X = X.loc[:, cur_view_features]
        selector1 = SelectPercentile(sym_auc_score,
                                     percentile=self.percentile)
        selector1.fit(cur_X, y)
        support = selector1.get_support()
        self.support[view] = support
        self.data[view] = {}
        self.data[view]['X'] = cur_X.loc[:, support]
        self.data[view]['y'] = y

    def multi_view_prefilter(self, X, y=None):
        non_target_views = self.views.copy()
        non_target_views.remove(self.target_view)
        # target view
        # find features in each view
        ptn = re.compile(r'^{}_'.format(self.target_view))
        cur_view_features = [i for i in self.all_features
                             if ptn.match(i)]
        cur_X = X.loc[:, cur_view_features]
        selector1 = SelectPercentile(sym_auc_score,
                                     percentile=self.percentile)
        selector1.fit(cur_X, y)
        # find the score cutoff
        scores = selector1.scores_
        cutoff = np.sort(scores[selector1.get_support()])[0]
        support = selector1.get_support()
        self.support[self.target_view] = support
        self.data[self.target_view] = {}
        self.data[self.target_view]['X'] = cur_X.loc[:, support]
        self.data[self.target_view]['y'] = y

        for view in non_target_views:
            # find features in each view
            ptn = re.compile(r'^{}_'.format(view))
            cur_view_features = [i for i in self.all_features
                                 if ptn.match(i)]
            cur_X = X.loc[:, cur_view_features]
            selector1 = SelectMinScore(score_func=sym_auc_score,
                                       min_score=cutoff)
            selector1.fit(cur_X, y)
            support = selector1.get_support()
            self.support[view] = support
            if support.sum() > 0:
                self.data[view] = {}
                self.data[view]['X'] = cur_X.loc[:, support]
                self.data[view]['y'] = y

        if len(self.data) == 1:
            warnings.warn('non target views contributed zero feature.')

    def pre_filter(self, X, y=None):
        ''' X is a combined multi-view data frame
            prefiltering for each view
        '''
        if self.mode == 'so':
            self.single_view_prefilter(X, y)
        else:
            self.multi_view_prefilter(X, y)

    def fit(self, X, y=None):
        self.all_features = X.columns.tolist()
        self.pre_filter(X, y)
        self.results = self.feature_sel()
        self.selected_features = self.results[0]
        if len(self.results) == 2:
            self.cluster_membership = self.results[1]
        return self

    def get_feature_names(self):
        return self.selected_features

    def get_cluster_membership(self):
        return self.cluster_membership

    def transform(self, X, y=None):
        if self.selected_features is not None:
            return X.loc[:, self.selected_features]

class FeatureSelBase(object):
    """Base class for feature selection method"""
    def __init__(self, method_type, all_view_data, target_view_name, k):
        # type can be 'sv' (single view) or 'mv' (multi-view)
        self.method_type = method_type
        self.all_view_data = all_view_data
        self.target_view_name = target_view_name
        self.k = k

    def check_enough_feature(self, X, k):
        len_features = len(X.columns.values)
        warn_msg = 'not enough features in the target view'
        if len_features <= k:
            warnings.warn(warn_msg)

    def compute_feature_score(self, X, y, score_func=sym_auc_score):
        score_func_ret = score_func(X, y)
        if isinstance(score_func_ret, (list, tuple)):
            scores_, pvalues_ = score_func_ret
            pvalues_ = np.asarray(pvalues_)
        else:
            scores_ = score_func_ret
            pvalues_ = None
        # for now ignore pvalues
        return scores_


class ProMS(FeatureSelBase):
    """ProMS single view"""
    def __init__(self, all_view_data, target_view_name, k,
                 weighted=True):
        self.weighted = weighted
        super().__init__('sv', all_view_data, target_view_name, k)

    def __call__(self):
        # use default parameters
        km = KMedoids(n_clusters=self.k, init='k-medoids++', max_iter=300,
                      metric='correlation', random_state=0)
        # feature wise k-medoids clustering
        X = self.all_view_data[self.target_view_name]['X']
        y = self.all_view_data[self.target_view_name]['y']
        self.check_enough_feature(X, self.k)

        if self.weighted:
            feature_scores = self.compute_feature_score(X, y)

        all_target_feature_names = X.columns.values
        X = X.T
        if self.weighted:
            km.fit(X, sample_weight=feature_scores)
        else:
            km.fit(X)
        km.fit(X)

        # the cluster centers are real data points
        selected_target_features_idx = km.medoid_indices_
        # print out the class membership
        cluster_label = km.labels_
        cluster_membership = dict()
        selected_features = []
        for i, _ in enumerate(selected_target_features_idx):
            cur_idx = selected_target_features_idx[i]
            cur_label = cluster_label[cur_idx]
            cur_selected = all_target_feature_names[cur_idx]
            selected_features.append(cur_selected)
            cluster_members = []
            indices = [j for j, x in enumerate(cluster_label)
                       if x == cur_label]
            # remove self
            indices.remove(cur_idx)
            if len(indices) > 0:
                cluster_members = [all_target_feature_names[j] for j in
                                   indices]
            else:
                cluster_members = []
            cluster_membership[cur_selected] = cluster_members
        return (selected_features, cluster_membership)


class ProMS_mo(FeatureSelBase):
    """Multiomics ProMS"""
    def __init__(self, all_view_data, target_view_name, k,
                 weighted=True):
        self.weighted = weighted
        super().__init__('mv', all_view_data, target_view_name, k)

    def __call__(self):
        km = KMedoids(n_clusters=self.k, init='k-medoids++', max_iter=300,
                      metric='correlation', random_state=0)
        all_X = None
        all_features = pd.DataFrame(columns=['name', 'view'])
        if self.weighted:
            feature_scores = np.array([], dtype=np.float64)

        candidacy = np.array([], dtype=np.bool)

        for i in self.all_view_data:
            X = self.all_view_data[i]['X']
            y = self.all_view_data[i]['y']
            if i == self.target_view_name:
                self.check_enough_feature(X, self.k)

            all_X = pd.concat([all_X, X], axis=1)
            cur_features = pd.DataFrame(columns=['name', 'view'])
            cur_features['name'] = X.columns.values
            cur_features['view'] = i
            all_features = pd.concat([all_features, cur_features], axis=0)
            candidacy = np.concatenate((candidacy, np.repeat(i ==
                                       self.target_view_name,
                                       len(cur_features.index))))
            if self.weighted:
                cur_feature_scores = self.compute_feature_score(X, y)
                feature_scores = np.concatenate((feature_scores,
                                                cur_feature_scores))

        all_feature_names = all_X.columns.values
        # feature wise k-medoids clustering
        # all_X is now of shape:  n_features x n_sample
        all_X = all_X.T
        if self.weighted:
            km.fit(all_X, sample_weight=feature_scores, candidacy=candidacy)
        else:
            km.fit(all_X, candidacy=candidacy)

        # the cluster centers are real data points
        selected_target_features_idx = km.medoid_indices_
        # print out the class membership
        cluster_label = km.labels_
        cluster_membership = dict()
        selected_features = []
        for i, _ in enumerate(selected_target_features_idx):
            cur_idx = selected_target_features_idx[i]
            cur_label = cluster_label[cur_idx]
            cur_selected = all_feature_names[cur_idx]
            selected_features.append(cur_selected)
            cluster_members = []
            indices = [j for j, x in enumerate(cluster_label)
                       if x == cur_label]
            # remove self
            indices.remove(cur_idx)
            if len(indices) > 0:
                cluster_members = [all_feature_names[j] for j in
                                   indices]
            else:
                cluster_members = []
            cluster_membership[cur_selected] = cluster_members
        return (selected_features, cluster_membership)


fs_methods = {
    'proms': ProMS,
    'proms_mo': ProMS_mo
}



import os
import json
import pandas as pd
import argparse
import pickle
from sklearn.model_selection import train_test_split
from proms import Data

class Dataset(Data):
    """ data set"""

    def __call__(self):
        print('processing data ...')
        predict_dataset = self.config['predict_dataset']

        # get the name of samples that have all the omics data available
        all_views = self.config['data'][predict_dataset]['view']
        n_views = len(all_views)
        all_view_names = [item['type'] for item in all_views]

        all_data_ = {'desc': {
                        'name': self.name,
                        'n_view': n_views,
                        'view_names': all_view_names
                        },
                     'X': {}
                    }

        for i in range(n_views):
            view_name = all_views[i]['type']
            print(f'view: {view_name}')
            view_file = os.path.join(self.root, predict_dataset,
                                     all_views[i]['file'])
            X = self.load_data(view_file, None, vis=False)

            X = X.add_prefix(view_name + '_')
            all_data_['X'][view_name] = X

        print('number of samples: {}'.format(all_data_['X'][view_name].shape[0]))

        self.all_data = all_data_
        return all_data_


def prepare_data(all_data):
    n_view = all_data['desc']['n_view']
    dataset_name = all_data['desc']['name']
    view_names = all_data['desc']['view_names']

    data = {'desc': {'view_names': view_names},
            'train': {}
            }
    data['desc']['name'] = dataset_name

    for i in range(n_view):
        cur_view_name = view_names[i]
        data['train'][cur_view_name] = {}
        cur_X = all_data['X'][cur_view_name]
        data['train'][cur_view_name]['X'] = cur_X

    data['train']['y'] = None

    return data


def create_dataset(config_file, output_run=None):
    """ create data structure from input data files """
    print(f'data config file: {config_file}')
    with open(config_file) as config_fh:
        data_config = json.load(config_fh)
        data_root = data_config['data_root']
        if not os.path.isabs(data_root):
            # relative to the data config file
            config_root = os.path.abspath(os.path.dirname(config_file))
            data_root = os.path.join(config_root, data_root)
        ds_name = data_config['name']
    all_data = Dataset(name=ds_name, root=data_root, config_file=config_file,
                       output_dir=output_run)()
    return all_data


def main():
    parser = get_parser()
    args = parser.parse_args()
    model_file = args.saved_model_file
    data_config_file = args.data_config_file

    with open(model_file, 'rb') as m_fh:
        saved_model = pickle.load(m_fh)

    all_data = create_dataset(data_config_file)
    data = prepare_data(all_data)

    X_train_combined = None
    view_names = data['desc']['view_names']
    for view in view_names:
        cur_train = data['train'][view]
        X_train_combined = pd.concat([X_train_combined, cur_train['X']],
                                     axis=1)
    pred_prob = saved_model.predict_proba(X_train_combined)[:,1]
    pred_label = [1 if x >= 0.5 else 0 for x in pred_prob]
    print(f'predicted probability: {pred_prob}')
    print(f'predicted label: {pred_label}')

def is_valid_file(arg):
    """ check if the file exists """
    arg = os.path.abspath(arg)
    if not os.path.exists(arg):
        msg = "The file %s does not exist!" % arg
        raise argparse.ArgumentTypeError(msg)
    else:
        return arg


def get_parser():
    ''' get arguments '''
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--model', dest='saved_model_file',
                        type=is_valid_file,
                        required=True,
                        help='saved model pickle file',
                        metavar='FILE',
                        )
    parser.add_argument('-d', '--data', dest='data_config_file',
                        type=is_valid_file,
                        required=True,
                        help='configuration file for data set',
                        metavar='FILE',
                        )
    return parser


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import os

__version__ = '1.0.4'
url = 'https://github.com/bzhanglab/proms'

install_requires = [
    'numpy',
    'scipy',
    'sklearn',
    'pandas',
    'xgboost',
    'matplotlib',
    
]


setup(
    name='proms',
    version=__version__,
    description='Protein Markers Selection',
    author='Zhiao Shi',
    author_email='zhiao.shi@gmail.com',
    url=url,
    download_url='{}/archive/{}.tar.gz'.format(url, __version__),
    keywords=[
        'machine learning', 'feature selection', 'proteomics',
        'multiomics', 'biomarker', 'genomics', 'bioinformatics'
    ],
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.6',
    install_requires=install_requires,
    packages=find_packages(),
    entry_points={"console_scripts": ["proms_train=proms.__main__:main", 
                  "proms_predict=proms.predict:main"]}
)



import multiprocessing
import os
import json
import csv
from datetime import datetime
from tempfile import mkdtemp
from shutil import rmtree
from numpy.lib.npyio import save
import pandas as pd
import argparse
import pickle
from xgboost import XGBClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import StratifiedKFold
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline
from joblib import Memory
from proms import Dataset, FeatureSelector


def run_single_fs_method(data, fs_method, run_config, k, repeat, estimator,
                         mode, seed):
    n_jobs = run_config['n_jobs']
    percentile = run_config['percentile']

    classifier_dict = {
        'xgboost': XGBClassifier(random_state=seed),
        'rf': RandomForestClassifier(random_state=seed),
        'svm': SVC(kernel='rbf', probability=True, random_state=seed),
        'logreg': LogisticRegression(max_iter=1000, random_state=seed)
    }
    classifier = classifier_dict[estimator]
    print(f'k={k}, repeat={repeat+1}, estimator={estimator}, fs_method={fs_method}', 
          flush=True)

    # pipeline steps
    target_view_name = data['desc']['target_view_name']
    X_train_combined = None
    y_train = data['train']['y']
    X_test_combined = None
    X_test_2 = None
    view_names = data['desc']['view_names']
    for view in view_names:
        cur_train = data['train'][view]
        X_train_combined = pd.concat([X_train_combined, cur_train['X']],
                                     axis=1)
        if data['test'] is not None:
            cur_test = data['test'][view]
            X_test_combined = pd.concat([X_test_combined, cur_test['X']],
                                        axis=1)
    if data['test'] is not None:
        y_test = data['test']['y']
    if data['test_2'] is not None:
        X_test_2 = data['test_2']['X']
        y_test_2 = data['test_2']['y']

    p_steps = []
    fs = FeatureSelector(views=view_names,
                         target_view=target_view_name,
                         method=fs_method,
                         k=k,
                         weighted=True)
    p_steps.append(('featureselector', fs))
    p_steps.append((estimator, classifier))
    cachedir = mkdtemp()
    memory = Memory(location=cachedir, verbose=0)
    pipe = Pipeline(steps=p_steps, memory=memory)
    param_grid = get_estimator_parameter_grid(estimator)
    param_grid['featureselector__percentile'] = percentile
    grid = GridSearchCV(pipe, param_grid, scoring='roc_auc',
                        cv=StratifiedKFold(3),
                        n_jobs=n_jobs,
                        verbose=1)
    grid.fit(X_train_combined, y_train.values.ravel())
    rmtree(cachedir)

    # evaluate on the test data set (if available)
    # and independent test set (if available)
    test_acc = test_score = 'NA'
    label = pred_label_s = pred_prob_s = 'NA'
    test_acc_2 = test_score_2 = 'NA'
    label_2 = pred_label_s_2 = pred_prob_s_2 = 'NA'
    if X_test_combined is not None:
        label = ','.join(map(str, y_test.to_numpy().squeeze()))
        pred_prob = grid.predict_proba(X_test_combined)[:, 1]
        pred_label = [1 if x >= 0.5 else 0 for x in pred_prob]
        test_score = grid.score(X_test_combined, y_test.values.ravel())
        pred_prob_s = ','.join(map('{:.4f}'.format, pred_prob))
        pred_label_s = ','.join(map(str, pred_label))
        test_acc = accuracy_score(y_test.to_numpy().squeeze(), pred_label)

    if X_test_2 is not None:
        pred_prob_2 = grid.predict_proba(X_test_2)[:, 1]
        pred_label_2 = [1 if x >= 0.5 else 0 for x in pred_prob_2]
        pred_prob_s_2 = ','.join(map('{:.4f}'.format, pred_prob_2))
        pred_label_s_2 = ','.join(map(str, pred_label_2))
        if y_test_2 is not None:
            label_2 = ','.join(map(str, y_test_2.to_numpy().squeeze()))
            test_score_2 = round(grid.score(X_test_2, y_test_2.values.ravel()), 4)
            test_acc_2 = round(accuracy_score(y_test_2.to_numpy().squeeze(),
                            pred_label_2), 4)

    # best_percentile = grid.best_params_['featureselector__percentile']
    best_fs = grid.best_estimator_.named_steps['featureselector']
    selected_features = best_fs.get_feature_names()

    if mode == 'full':
        cluster_membership = best_fs.get_cluster_membership()
        run_version = run_config['run_version']
        output_root = run_config['output_root']
        out_dir_run = os.path.join(output_root, run_version)
        out_dir_run_full = os.path.join(out_dir_run, 'full_model')
        with open(os.path.join(out_dir_run_full, 'full_model.pkl'), 'wb') as fh:
            pickle.dump(grid.best_estimator_, fh, protocol=pickle.HIGHEST_PROTOCOL)

    omics_type = 'so' if fs_method == 'proms' else 'mo'
    if mode == 'eval':
        res = [fs_method, omics_type, k,
               estimator, repeat, pred_prob_s,
               pred_label_s, label, round(test_acc, 4), round(test_score, 4)]
        print(f'acc:{round(test_acc, 4)}, auroc:{round(test_score, 4)}')
        return res
    elif mode == 'full':
        s_features = ','.join(selected_features)
        if data['test_2']['X'] is not None:
            if data['test_2']['y'] is not None:
                res = [fs_method, omics_type, k, estimator,
                    s_features, json.dumps(cluster_membership), -1, -1,
                    pred_prob_s_2, pred_label_s_2,
                    label_2, test_acc_2, test_score_2]
                print(f'acc:{test_acc_2}, auroc:{test_score_2}')
            else:
                res = [fs_method, omics_type, k, estimator,
                    s_features, json.dumps(cluster_membership), -1, -1,
                    pred_prob_s_2, pred_label_s_2]
        else:
            res = [fs_method, omics_type, k, estimator,
                s_features, json.dumps(cluster_membership), -1, -1]

        return res


def run_single_estimator(data, run_config, k, repeat, fs_method,
                         estimator, mode, seed):
    n_view = len(data['desc']['view_names'])
    res = []
    if fs_method is None:
        # proms and proms_mo (if there are more than 1 view available)
        cur_res = run_single_fs_method(data, 'proms', run_config, k, repeat,
                      estimator, mode, seed)
    
        res.append(cur_res)
        if n_view > 1:
            cur_res = run_single_fs_method(data, 'proms_mo', run_config, k,
                          repeat, estimator, mode, seed)
            res.append(cur_res)
    else:
        cur_res = run_single_fs_method(data, fs_method, run_config, k, repeat,
                      estimator, mode, seed)
        res.append(cur_res)

    return res


def prepare_data(all_data, repeat, mode):
    test_ratio = 0.3
    y = all_data['y']
    n_view = all_data['desc']['n_view']
    dataset_name = all_data['desc']['name']
    view_names = all_data['desc']['view_names']
    target_view_name = all_data['desc']['target_view']
    if mode == 'eval':
        y_train, y_test = train_test_split(y, test_size=test_ratio,
                                           stratify=y, random_state=repeat)
    else:
        y_train, y_test = y, None

    data = {'desc': {},
            'train': {},
            'test': None,
            'test_2': None
            }
    data['desc']['name'] = dataset_name
    data['desc']['view_names'] = view_names
    data['desc']['target_view_name'] = target_view_name
    if mode == 'eval':
        data['test'] = {}
    if 'X_test' in all_data:
        data['test_2'] = {}

    for i in range(n_view):
        cur_view_name = view_names[i]
        data['train'][cur_view_name] = {}
        if mode == 'eval':
            data['test'][cur_view_name] = {}
        cur_X = all_data['X'][cur_view_name]
        cur_X_train = cur_X.loc[y_train.index, :]
        data['train'][cur_view_name]['X'] = cur_X_train
        if mode == 'eval':
            cur_X_test = cur_X.loc[y_test.index, :]
            data['test'][cur_view_name]['X'] = cur_X_test

        if 'X_test' in all_data:
            data['test_2']['X'] = all_data['X_test']
            data['test_2']['y'] = all_data['y_test']

    data['train']['y'] = y_train
    if mode == 'eval':
        data['test']['y'] = y_test

    return data


def run_single_repeat(all_data, run_config, k, repeat, fs_method,
                      mode, seed):
    data = prepare_data(all_data, repeat, mode)
    classifiers = run_config['classifiers']
    res = []
    for estimator in classifiers:
        cur_res = run_single_estimator(data, run_config, k, repeat, fs_method,
                      estimator, mode, seed)
        res.extend(cur_res)
    return res


def run_single_k(all_data, run_config, k, fs_method, mode, seed):
    n_repeats = run_config['repeat']
    res = []
    for repeat in range(n_repeats):
        cur_res = run_single_repeat(all_data, run_config, k, repeat,
                      fs_method, mode, seed)
        res.extend(cur_res)
    return res


def get_estimator_parameter_grid(classifier):
    if classifier == 'xgboost':
        param_grid = {classifier + '__max_depth': [2, 4, 6, 8, 10],
                      classifier + '__n_estimators': range(50, 400, 50)
                      }
    elif classifier == 'rf':
        param_grid = {classifier + '__max_depth': [2, 4, 6, 8, 10],
                      classifier + '__n_estimators': range(50, 400, 50)
                      }
    elif classifier == 'svm':
        param_grid = {classifier + '__C': [0.001, 0.01, 0.1, 1, 10, 100],
                      classifier + '__gamma': [0.001, 0.01, 0.1, 1, 10, 100]
                      }
    elif classifier == 'logreg':
        param_grid = {classifier + '__C': [0.001, 0.01, 0.1, 1, 10, 100],
                      }
    else:
        raise ValueError(f'classifier {classifier} not supported')
    return param_grid


def run_fs(all_data, run_config, run_version, output_root, seed):
    k = run_config['k']
    dataset_name = all_data['desc']['name']

    # evaluate: performance evaluation, select features with
    #           model built with train set and evaluate in validation set
    column_names = ['fs', 'type', 'k', 'classifier', 'repeat',
                    'val_score', 'val_pred_label',
                    'val_label', 'val_acc', 'val_auroc']
    out_dir_run = os.path.join(output_root, run_version)
    res = []
    for cur_k in k:
        cur_res = run_single_k(all_data, run_config, cur_k, None, 'eval', seed)
        res.extend(cur_res)

    results_df = pd.DataFrame(res, columns=column_names)
    out_file = dataset_name + '_results_'
    out_file = out_file + run_version + '_eval.tsv'
    out_file = os.path.join(out_dir_run, out_file)
    results_df.to_csv(out_file, header=True, sep='\t', index=False)

    # we will select the best combination of fs, k, classifier
    # based on cross validation results (average aurco) from the previous step
    # to fit a full model
   
    results_df_sel = results_df[['fs', 'k', 'classifier', 'val_acc',
                                'val_auroc']]
    results_df_sel = results_df_sel.groupby(['fs', 'k', 'classifier']).mean()
    results_df_sel = results_df_sel.reset_index()
    best_auroc_idx = results_df_sel['val_auroc'].argmax()
    fs_sel = results_df_sel.loc[best_auroc_idx, 'fs']
    k_sel = results_df_sel.loc[best_auroc_idx, 'k']
    classifier_sel = results_df_sel.loc[best_auroc_idx, 'classifier']
    best_avg_acc = results_df_sel.loc[best_auroc_idx, 'val_acc']
    best_avg_auroc = results_df_sel.loc[best_auroc_idx, 'val_auroc']

    # # full: build full model with all data and test in another
    # #       indpendent data set if available
    if all_data['desc']['has_test']:
        if all_data['y_test'] is not None:
            column_names = ['fs', 'type', 'k', 'classifier',
                            'features', 'membership',
                            'avg_val_acc', 'avg_val_auroc',
                            'test_score', 'test_pred_label', 'test_label',
                            'test_accuracy', 'test_auroc']
        else:
            column_names = ['fs', 'type', 'k', 'classifier',
                            'features', 'membership',
                            'avg_val_acc', 'avg_val_auroc',
                            'test_score', 'test_pred_label']
    else:
        column_names = ['fs', 'type', 'k', 'classifier',
                        'features', 'membership',
                        'avg_val_acc', 'avg_val_auroc']
    # re-define run_config file
    run_config['repeat'] = 1
    run_config['k'] = [k_sel]
    run_config['classifiers'] = [classifier_sel]
    fs_method = fs_sel

    res = []
    cur_res = run_single_k(all_data, run_config, k_sel, fs_method,
                           'full', seed)
    cur_res[0][6] = best_avg_acc
    cur_res[0][7] = best_avg_auroc
    res.extend(cur_res)
    results_df = pd.DataFrame(res, columns=column_names)
    out_file = dataset_name + '_results_'
    out_file = out_file + run_version + '_full.tsv'
    out_file = os.path.join(out_dir_run, out_file)
    results_df.to_csv(out_file, header=True, sep='\t', index=False, quoting=csv.QUOTE_NONE)


def check_data_config(config_file):
    with open(config_file) as config_fh:
        data_config = json.load(config_fh)

    required_fields = {'name', 'data_root', 'train_dataset', 'target_view',
                       'target_label', 'data'}
    allowed_fields = {'test_dataset'} | required_fields
    if not required_fields <= data_config.keys() <= allowed_fields:
        raise Exception(f'config required fields: {required_fields}, allowed fields: {allowed_fields}')

    data_required_fields = {'train'}
    data_allowed_fields = {'test'} | data_required_fields
    if not data_required_fields <= data_config['data'].keys() <= data_allowed_fields:
        raise Exception(f'data required fields: {data_required_fields}, allowed fileds: {data_allowed_fields}')

    train_required_fields = {'label', 'view'}
    if not train_required_fields <= data_config['data']['train'].keys():
        raise Exception(f'train data required fields: {train_required_fields}')


def create_dataset(config_file, output_run):
    """ create data structure from input data files """
    print(f'data config file: {config_file}')
    check_data_config(config_file)
    with open(config_file) as config_fh:
        data_config = json.load(config_fh)
        data_root = data_config['data_root']
        if not os.path.isabs(data_root):
            # relative to the data config file
            config_root = os.path.abspath(os.path.dirname(config_file))
            data_root = os.path.join(config_root, data_root)
        ds_name = data_config['name']
    all_data = Dataset(name=ds_name, root=data_root, config_file=config_file,
                       output_dir=output_run)()
    return all_data


def check_run_config(run_config_file, n_train_sample):
    with open(run_config_file) as config_fh:
        run_config = json.load(config_fh)
        if not 'n_jobs' in run_config:
            # assume running on a node with 4 cpus
            run_config['n_jobs'] = 4
        if not 'repeat' in run_config:
            run_config['repeat'] = 5
        if not 'k' in run_config:
            raise Exception('must specifiy k in run configuration file.')
        else:
            k = run_config['k']
            k_max = sorted(k)[-1]
            if k_max > int(0.25*n_train_sample):
                raise Exception('largest k should be less than 25% of the number of training samples')
        default_classifiers = ['logreg', 'rf', 'svm', 'xgboost']
        if not 'classifiers' in run_config:
            run_config['classifiers'] = default_classifiers
        else:
            if not set(run_config['classifiers']).issubset(default_classifiers):
                raise Exception(f'supported classifiers:{default_classifiers}')
        if not 'percentile' in run_config:
            run_config['percentile'] = [1, 5, 10, 20]
        else:
            all_ok = all(x > 0 and x < 50 for x in run_config['percentile']) 
            if not all_ok: 
                raise Exception('all percentile values must be > 0 and < 50')
    
    return run_config


def main():
    parser = get_parser()
    args = parser.parse_args()
    run_config_file = args.run_config_file
    data_config_file = args.data_config_file
    output_root = args.output_root
    # time stamp
    run_version = args.run_version
    # random seed for full model (if applicable)
    seed = args.seed
    # prepare output directory
    out_dir_run = os.path.join(output_root, run_version)
    out_dir_run_full = os.path.join(out_dir_run, 'full_model')
    if not os.path.exists(out_dir_run_full):
        os.makedirs(out_dir_run_full)
    all_data = create_dataset(data_config_file, out_dir_run)
    n_train_sample = all_data['y'].shape[0]
    run_config = check_run_config(run_config_file, n_train_sample)
    run_config['run_version'] = run_version
    run_config['output_root'] = output_root

    run_fs(all_data, run_config, run_version, output_root, seed)


def is_valid_file(arg):
    """ check if the file exists """
    arg = os.path.abspath(arg)
    if not os.path.exists(arg):
        msg = "The file %s does not exist!" % arg
        raise argparse.ArgumentTypeError(msg)
    else:
        return arg


def date_time_now():
    """ get the current date and time """
    now = datetime.now()
    date_time = now.strftime("%Y_%m_%d_%H_%M_%S")
    return date_time


def get_parser():
    ''' get arguments '''
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file', dest='run_config_file',
                        type=is_valid_file,
                        required=True,
                        help='configuration file for the run',
                        metavar='FILE',
                        )
    parser.add_argument('-d', '--data', dest='data_config_file',
                        type=is_valid_file,
                        required=True,
                        help='configuration file for data set',
                        metavar='FILE',
                        )
    parser.add_argument('-s', '--seed', dest='seed',
                        default=42,
                        type=int,
                        help='random seed '
                        )
    parser.add_argument('-o', '--output', dest='output_root',
                        default='results',
                        type=str,
                        help='output directory'
                        )
    parser.add_argument('-r', '--run_version', dest='run_version',
                        default=date_time_now(),
                        type=str,
                        help='name of the run, default to current date/time'
                        )
    return parser


if __name__ == '__main__':
    multiprocessing.set_start_method('forkserver')
    main()
